@@ -30,7 +30,7 @@ class TelemetryEvent(Enum):
 
     @classmethod
     def can_share(cls) -> List["TelemetryEvent"]:
-        """ only these events will be shared to the central telemetry """
+        """only these events will be shared to the central telemetry"""
         return [cls.task, cls.state_changed]
 
 
@@ -44,7 +44,7 @@ class TelemetryData(Enum):
 
     @classmethod
     def can_share(cls) -> List["TelemetryData"]:
-        """ only these types of data will be shared to the central telemetry """
+        """only these types of data will be shared to the central telemetry"""
         return [cls.current_state, cls.vm_id, cls.job_id, cls.task_id, cls.task_type]
 
 
@@ -57,6 +57,7 @@ class TaskFeature(Enum):
     stats_file = "stats_file"
     stats_format = "stats_format"
     target_exe = "target_exe"
+    target_exe_optional = "target_exe_optional"
     target_env = "target_env"
     target_options = "target_options"
     analyzer_exe = "analyzer_exe"
@@ -74,6 +75,12 @@ class TaskFeature(Enum):
     check_debugger = "check_debugger"
     check_retry_count = "check_retry_count"
     ensemble_sync_delay = "ensemble_sync_delay"
+    preserve_existing_outputs = "preserve_existing_outputs"
+    check_fuzzer_help = "check_fuzzer_help"
+    expect_crash_on_failure = "expect_crash_on_failure"
+    report_list = "report_list"
+    minimized_stack_depth = "minimized_stack_depth"
+    coverage_filter = "coverage_filter"
 
 
 # Permissions for an Azure Blob Storage Container.
@@ -82,10 +89,8 @@ class TaskFeature(Enum):
 class ContainerPermission(Enum):
     Read = "Read"
     Write = "Write"
-    Create = "Create"
     List = "List"
     Delete = "Delete"
-    Add = "Add"
 
 
 class JobState(Enum):
@@ -96,7 +101,7 @@ class JobState(Enum):
 
     @classmethod
     def available(cls) -> List["JobState"]:
-        """ set of states that indicate if tasks can be added to it """
+        """set of states that indicate if tasks can be added to it"""
         return [x for x in cls if x not in [cls.stopping, cls.stopped]]
 
     @classmethod
@@ -106,6 +111,10 @@ class JobState(Enum):
         """
         return [cls.init, cls.stopping]
 
+    @classmethod
+    def shutting_down(cls) -> List["JobState"]:
+        return [cls.stopping, cls.stopped]
+
 
 class TaskState(Enum):
     init = "init"
@@ -130,7 +139,7 @@ class TaskState(Enum):
 
     @classmethod
     def available(cls) -> List["TaskState"]:
-        """ set of states that indicate if the task isn't stopping """
+        """set of states that indicate if the task isn't stopping"""
         return [x for x in cls if x not in [TaskState.stopping, TaskState.stopped]]
 
     @classmethod
@@ -139,15 +148,18 @@ class TaskState(Enum):
 
 
 class TaskType(Enum):
+    coverage = "coverage"
     libfuzzer_fuzz = "libfuzzer_fuzz"
     libfuzzer_coverage = "libfuzzer_coverage"
     libfuzzer_crash_report = "libfuzzer_crash_report"
     libfuzzer_merge = "libfuzzer_merge"
+    libfuzzer_regression = "libfuzzer_regression"
     generic_analysis = "generic_analysis"
     generic_supervisor = "generic_supervisor"
     generic_merge = "generic_merge"
     generic_generator = "generic_generator"
     generic_crash_report = "generic_crash_report"
+    generic_regression = "generic_regression"
 
 
 class VmState(Enum):
@@ -168,7 +180,7 @@ class VmState(Enum):
 
     @classmethod
     def available(cls) -> List["VmState"]:
-        """ set of states that indicate if the repro vm isn't stopping """
+        """set of states that indicate if the repro vm isn't stopping"""
         return [x for x in cls if x not in [cls.stopping, cls.stopped]]
 
 
@@ -201,6 +213,7 @@ class ContainerType(Enum):
     tools = "tools"
     unique_inputs = "unique_inputs"
     unique_reports = "unique_reports"
+    regression_reports = "regression_reports"
 
     @classmethod
     def reset_defaults(cls) -> List["ContainerType"]:
@@ -213,10 +226,15 @@ class ContainerType(Enum):
             cls.readonly_inputs,
             cls.reports,
             cls.setup,
-            cls.unique_reports,
             cls.unique_inputs,
+            cls.unique_reports,
+            cls.regression_reports,
         ]
 
+    @classmethod
+    def user_config(cls) -> List["ContainerType"]:
+        return [cls.setup, cls.inputs, cls.readonly_inputs]
+
 
 class StatsFormat(Enum):
     AFL = "AFL"
@@ -244,6 +262,8 @@ class ErrorCode(Enum):
     TASK_FAILED = 468
     INVALID_NODE = 469
     NOTIFICATION_FAILURE = 470
+    UNABLE_TO_UPDATE = 471
+    PROXY_FAILED = 472
 
 
 class HeartbeatType(Enum):
@@ -271,7 +291,7 @@ class PoolState(Enum):
 
     @classmethod
     def available(cls) -> List["PoolState"]:
-        """ set of states that indicate if it's available for work """
+        """set of states that indicate if it's available for work"""
         return [cls.running]
 
 
@@ -284,6 +304,13 @@ class ScalesetState(Enum):
     halt = "halt"
     creation_failed = "creation_failed"
 
+    @classmethod
+    def can_update(cls) -> List["ScalesetState"]:
+        """
+        set of states that indicate the scaleset can be updated
+        """
+        return [cls.running, cls.resize]
+
     @classmethod
     def needs_work(cls) -> List["ScalesetState"]:
         """
@@ -293,13 +320,13 @@ class ScalesetState(Enum):
 
     @classmethod
     def available(cls) -> List["ScalesetState"]:
-        """ set of states that indicate if it's available for work """
+        """set of states that indicate if it's available for work"""
         unavailable = [cls.shutdown, cls.halt, cls.creation_failed]
         return [x for x in cls if x not in unavailable]
 
     @classmethod
     def modifying(cls) -> List["ScalesetState"]:
-        """ set of states that indicate scaleset is resizing """
+        """set of states that indicate scaleset is resizing"""
         return [
             cls.halt,
             cls.init,
@@ -344,6 +371,10 @@ class NodeState(Enum):
         # from the agent.
         return [cls.done, cls.shutdown, cls.halt]
 
+    @classmethod
+    def can_process_new_work(cls) -> List["NodeState"]:
+        return [cls.free]
+
 
 class GithubIssueState(Enum):
     open = "open"
@@ -360,15 +391,21 @@ class TaskDebugFlag(Enum):
     keep_node_on_completion = "keep_node_on_completion"
 
 
-class WebhookEventType(Enum):
-    task_created = "task_created"
-    task_stopped = "task_stopped"
-    task_failed = "task_failed"
-    ping = "ping"
-
-
 class WebhookMessageState(Enum):
     queued = "queued"
     retrying = "retrying"
     succeeded = "succeeded"
     failed = "failed"
+
+
+class UserFieldOperation(Enum):
+    add = "add"
+    replace = "replace"
+
+
+class UserFieldType(Enum):
+    Bool = "Bool"
+    Int = "Int"
+    Str = "Str"
+    DictStr = "DictStr"
+    ListStr = "ListStr"
