@@ -4,11 +4,13 @@
 # Licensed under the MIT License.
 
 from datetime import datetime
-from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union
+from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union
 from uuid import UUID, uuid4
 
 from pydantic import BaseModel, Field, root_validator, validator
+from pydantic.dataclasses import dataclass
 
+from ._monkeypatch import _check_hotfix
 from .consts import ONE_HOUR, SEVEN_DAYS
 from .enums import (
     OS,
@@ -36,11 +38,44 @@ from .primitives import Container, PoolName, Region
 
 
 class UserInfo(BaseModel):
-    application_id: UUID
+    application_id: Optional[UUID]
     object_id: Optional[UUID]
     upn: Optional[str]
 
 
+# Stores the address of a secret
+class SecretAddress(BaseModel):
+    # keyvault address of a secret
+    url: str
+
+
+T = TypeVar("T")
+
+
+# This class allows us to store some data that are intended to be secret
+# The secret field stores either the raw data or the address of that data
+# This class allows us to maintain backward compatibility with existing
+# NotificationTemplate classes
+@dataclass
+class SecretData(Generic[T]):
+    secret: Union[T, SecretAddress]
+
+    def __init__(self, secret: Union[T, SecretAddress]):
+        try:
+            self.secret = SecretAddress.parse_obj(secret)
+        except Exception:
+            self.secret = secret
+
+    def __str__(self) -> str:
+        return self.__repr__()
+
+    def __repr__(self) -> str:
+        if isinstance(self.secret, SecretAddress):
+            return str(self.secret)
+        else:
+            return "[REDACTED]"
+
+
 class EnumModel(BaseModel):
     @root_validator(pre=True)
     def exactly_one(cls: Any, values: Any) -> Any:
@@ -84,38 +119,28 @@ class JobConfig(BaseModel):
     project: str
     name: str
     build: str
-    duration: int
-
-    @validator("duration", allow_reuse=True)
-    def check_duration(cls, value: int) -> int:
-        if value < ONE_HOUR or value > SEVEN_DAYS:
-            raise ValueError("invalid duration")
-        return value
+    duration: int = Field(ge=ONE_HOUR, le=SEVEN_DAYS)
 
 
 class ReproConfig(BaseModel):
     container: Container
     path: str
-    duration: int
-
-    @validator("duration", allow_reuse=True)
-    def check_duration(cls, value: int) -> int:
-        if value < ONE_HOUR or value > SEVEN_DAYS:
-            raise ValueError("invalid duration")
-        return value
+    duration: int = Field(ge=ONE_HOUR, le=SEVEN_DAYS)
 
 
 class TaskDetails(BaseModel):
     type: TaskType
-    duration: int
-    target_exe: str
-    target_env: Dict[str, str]
-    target_options: List[str]
+    duration: int = Field(ge=ONE_HOUR, le=SEVEN_DAYS)
+    target_exe: Optional[str]
+    target_env: Optional[Dict[str, str]]
+    target_options: Optional[List[str]]
     target_workers: Optional[int]
     target_options_merge: Optional[bool]
     check_asan_log: Optional[bool]
     check_debugger: Optional[bool] = Field(default=True)
-    check_retry_count: Optional[int]
+    check_retry_count: Optional[int] = Field(ge=0)
+    check_fuzzer_help: Optional[bool]
+    expect_crash_on_failure: Optional[bool]
     rename_output: Optional[bool]
     supervisor_exe: Optional[str]
     supervisor_env: Optional[Dict[str, str]]
@@ -131,28 +156,12 @@ class TaskDetails(BaseModel):
     stats_file: Optional[str]
     stats_format: Optional[StatsFormat]
     reboot_after_setup: Optional[bool]
-    target_timeout: Optional[int]
+    target_timeout: Optional[int] = Field(ge=1)
     ensemble_sync_delay: Optional[int]
-
-    @validator("check_retry_count", allow_reuse=True)
-    def validate_check_retry_count(cls, value: int) -> int:
-        if value is not None:
-            if value < 0:
-                raise ValueError("invalid check_retry_count")
-        return value
-
-    @validator("target_timeout", allow_reuse=True)
-    def check_target_timeout(cls, value: Optional[int]) -> Optional[int]:
-        if value is not None:
-            if value < 1:
-                raise ValueError("invalid target_timeout")
-        return value
-
-    @validator("duration", allow_reuse=True)
-    def check_duration(cls, value: int) -> int:
-        if value < ONE_HOUR or value > SEVEN_DAYS:
-            raise ValueError("invalid duration")
-        return value
+    preserve_existing_outputs: Optional[bool]
+    report_list: Optional[List[str]]
+    minimized_stack_depth: Optional[int]
+    coverage_filter: Optional[str]
 
 
 class TaskPool(BaseModel):
@@ -164,16 +173,10 @@ class TaskVm(BaseModel):
     region: Region
     sku: str
     image: str
-    count: int = Field(default=1)
+    count: int = Field(default=1, ge=0)
     spot_instances: bool = Field(default=False)
     reboot_after_setup: Optional[bool]
 
-    @validator("count", allow_reuse=True)
-    def check_count(cls, value: int) -> int:
-        if value <= 0:
-            raise ValueError("invalid count")
-        return value
-
 
 class TaskContainers(BaseModel):
     type: ContainerType
@@ -189,6 +192,7 @@ class TaskConfig(BaseModel):
     containers: List[TaskContainers]
     tags: Dict[str, str]
     debug: Optional[List[TaskDebugFlag]]
+    colocate: Optional[bool]
 
 
 class BlobRef(BaseModel):
@@ -199,7 +203,7 @@ class BlobRef(BaseModel):
 
 class Report(BaseModel):
     input_url: Optional[str]
-    input_blob: BlobRef
+    input_blob: Optional[BlobRef]
     executable: str
     crash_type: str
     crash_site: str
@@ -209,6 +213,34 @@ class Report(BaseModel):
     asan_log: Optional[str]
     task_id: UUID
     job_id: UUID
+    scariness_score: Optional[int]
+    scariness_description: Optional[str]
+    minimized_stack: Optional[List[str]]
+    minimized_stack_sha256: Optional[str]
+    minimized_stack_function_names: Optional[List[str]]
+    minimized_stack_function_names_sha256: Optional[str]
+    minimized_stack_function_lines: Optional[List[str]]
+    minimized_stack_function_lines_sha256: Optional[str]
+
+
+class NoReproReport(BaseModel):
+    input_sha256: str
+    input_blob: Optional[BlobRef]
+    executable: str
+    task_id: UUID
+    job_id: UUID
+    tries: int
+    error: Optional[str]
+
+
+class CrashTestResult(BaseModel):
+    crash_report: Optional[Report]
+    no_repro: Optional[NoReproReport]
+
+
+class RegressionReport(BaseModel):
+    crash_test_result: CrashTestResult
+    original_crash_test_result: Optional[CrashTestResult]
 
 
 class ADODuplicateTemplate(BaseModel):
@@ -220,7 +252,7 @@ class ADODuplicateTemplate(BaseModel):
 
 class ADOTemplate(BaseModel):
     base_url: str
-    auth_token: str
+    auth_token: SecretData[str]
     project: str
     type: str
     unique_fields: List[str]
@@ -228,15 +260,33 @@ class ADOTemplate(BaseModel):
     ado_fields: Dict[str, str]
     on_duplicate: ADODuplicateTemplate
 
-    def redact(self) -> None:
-        self.auth_token = "***"
+    # validator needed for backward compatibility
+    @validator("auth_token", pre=True, always=True)
+    def validate_auth_token(cls, v: Any) -> SecretData:
+        if isinstance(v, str):
+            return SecretData(secret=v)
+        elif isinstance(v, SecretData):
+            return v
+        elif isinstance(v, dict):
+            return SecretData(secret=v["secret"])
+        else:
+            raise TypeError(f"invalid datatype {type(v)}")
 
 
 class TeamsTemplate(BaseModel):
-    url: str
+    url: SecretData[str]
 
-    def redact(self) -> None:
-        self.url = "***"
+    # validator needed for backward compatibility
+    @validator("url", pre=True, always=True)
+    def validate_url(cls, v: Any) -> SecretData:
+        if isinstance(v, str):
+            return SecretData(secret=v)
+        elif isinstance(v, SecretData):
+            return v
+        elif isinstance(v, dict):
+            return SecretData(secret=v["secret"])
+        else:
+            raise TypeError(f"invalid datatype {type(v)}")
 
 
 class ContainerDefinition(BaseModel):
@@ -276,10 +326,11 @@ class ClientCredentials(BaseModel):
 class AgentConfig(BaseModel):
     client_credentials: Optional[ClientCredentials]
     onefuzz_url: str
-    pool_name: str
+    pool_name: PoolName
     heartbeat_queue: Optional[str]
-    instrumentation_key: Optional[str]
-    telemetry_key: Optional[str]
+    instance_telemetry_key: Optional[str]
+    microsoft_telemetry_key: Optional[str]
+    multi_tenant_domain: Optional[str]
     instance_id: UUID
 
 
@@ -288,11 +339,10 @@ class TaskUnitConfig(BaseModel):
     job_id: UUID
     task_id: UUID
     task_type: TaskType
-    instrumentation_key: Optional[str]
-    telemetry_key: Optional[str]
+    instance_telemetry_key: Optional[str]
+    microsoft_telemetry_key: Optional[str]
     heartbeat_queue: str
     # command_queue: str
-    back_channel_address: str
     input_queue: Optional[str]
     supervisor_exe: Optional[str]
     supervisor_env: Optional[Dict[str, str]]
@@ -307,6 +357,8 @@ class TaskUnitConfig(BaseModel):
     check_asan_log: Optional[bool]
     check_debugger: Optional[bool]
     check_retry_count: Optional[int]
+    check_fuzzer_help: Optional[bool]
+    expect_crash_on_failure: Optional[bool]
     rename_output: Optional[bool]
     generator_exe: Optional[str]
     generator_env: Optional[Dict[str, str]]
@@ -318,6 +370,9 @@ class TaskUnitConfig(BaseModel):
     stats_file: Optional[str]
     stats_format: Optional[StatsFormat]
     ensemble_sync_delay: Optional[int]
+    report_list: Optional[List[str]]
+    minimized_stack_depth: Optional[int]
+    coverage_filter: Optional[str]
 
     # from here forwards are Container definitions.  These need to be inline
     # with TaskDefinitions and ContainerTypes
@@ -331,6 +386,7 @@ class TaskUnitConfig(BaseModel):
     tools: CONTAINER_DEF
     unique_inputs: CONTAINER_DEF
     unique_reports: CONTAINER_DEF
+    regression_reports: CONTAINER_DEF
 
 
 class Forward(BaseModel):
@@ -343,11 +399,16 @@ class ProxyConfig(BaseModel):
     url: str
     notification: str
     region: Region
+    proxy_id: UUID
     forwards: List[Forward]
+    instance_telemetry_key: Optional[str]
+    microsoft_telemetry_key: Optional[str]
+    instance_id: UUID
 
 
 class ProxyHeartbeat(BaseModel):
     region: Region
+    proxy_id: UUID
     forwards: List[Forward]
     timestamp: datetime = Field(default_factory=datetime.utcnow)
 
@@ -396,30 +457,53 @@ class GithubIssueSearch(BaseModel):
 
 
 class GithubAuth(BaseModel):
-    user: str
-    personal_access_token: str
+    user: str = Field(min_length=1)
+    personal_access_token: str = Field(min_length=1)
 
 
 class GithubIssueTemplate(BaseModel):
-    auth: GithubAuth
-    organization: str
-    repository: str
-    title: str
+    auth: SecretData[GithubAuth]
+    organization: str = Field(min_length=1)
+    repository: str = Field(min_length=1)
+    title: str = Field(min_length=1)
     body: str
     unique_search: GithubIssueSearch
     assignees: List[str]
     labels: List[str]
     on_duplicate: GithubIssueDuplicate
 
-    def redact(self) -> None:
-        self.auth.user = "***"
-        self.auth.personal_access_token = "***"
+    # validator needed for backward compatibility
+    @validator("auth", pre=True, always=True)
+    def validate_auth(cls, v: Any) -> SecretData:
+        def try_parse_GithubAuth(x: dict) -> Optional[GithubAuth]:
+            try:
+                return GithubAuth.parse_obj(x)
+            except Exception:
+                return None
+
+        if isinstance(v, GithubAuth):
+            return SecretData(secret=v)
+        elif isinstance(v, SecretData):
+            return v
+        elif isinstance(v, dict):
+            githubAuth = try_parse_GithubAuth(v)
+            if githubAuth:
+                return SecretData(secret=githubAuth)
+
+            githubAuth = try_parse_GithubAuth(v["secret"])
+            if githubAuth:
+                return SecretData(secret=githubAuth)
+
+            return SecretData(secret=v["secret"])
+        else:
+            raise TypeError(f"invalid datatype {type(v)}")
 
 
 NotificationTemplate = Union[ADOTemplate, TeamsTemplate, GithubIssueTemplate]
 
 
 class Notification(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     container: Container
     notification_id: UUID = Field(default_factory=uuid4)
     config: NotificationTemplate
@@ -432,16 +516,19 @@ class JobTaskInfo(BaseModel):
 
 
 class Job(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     job_id: UUID = Field(default_factory=uuid4)
     state: JobState = Field(default=JobState.init)
     config: JobConfig
     error: Optional[str]
     end_time: Optional[datetime] = None
     task_info: Optional[List[JobTaskInfo]]
+    user_info: Optional[UserInfo]
 
 
 class TaskHeartbeatEntry(BaseModel):
     task_id: UUID
+    job_id: Optional[UUID]
     machine_id: UUID
     data: List[Dict[str, HeartbeatType]]
 
@@ -451,12 +538,43 @@ class NodeHeartbeatEntry(BaseModel):
     data: List[Dict[str, HeartbeatType]]
 
 
+class NodeCommandStopIfFree(BaseModel):
+    pass
+
+
+class StopNodeCommand(BaseModel):
+    pass
+
+
+class StopTaskNodeCommand(BaseModel):
+    task_id: UUID
+
+
+class NodeCommandAddSshKey(BaseModel):
+    public_key: str
+
+
+class NodeCommand(EnumModel):
+    stop: Optional[StopNodeCommand]
+    stop_task: Optional[StopTaskNodeCommand]
+    add_ssh_key: Optional[NodeCommandAddSshKey]
+    stop_if_free: Optional[NodeCommandStopIfFree]
+
+
+class NodeCommandEnvelope(BaseModel):
+    command: NodeCommand
+    message_id: str
+
+
 class Node(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     pool_name: PoolName
+    pool_id: Optional[UUID]
     machine_id: UUID
     state: NodeState = Field(default=NodeState.init)
     scaleset_id: Optional[UUID] = None
     tasks: Optional[List[Tuple[UUID, NodeTaskState]]] = None
+    messages: Optional[List[NodeCommand]] = None
     heartbeat: Optional[datetime]
     version: str = Field(default="1.0.0")
     reimage_requested: bool = Field(default=False)
@@ -477,43 +595,23 @@ class NodeTasks(BaseModel):
 
 class AutoScaleConfig(BaseModel):
     image: str
-    max_size: Optional[int]  # max size of pool
-    min_size: int = Field(default=0)  # min size of pool
+    max_size: int = Field(default=1000, le=1000, ge=0)  # max size of pool
+    min_size: int = Field(default=0, le=1000, ge=0)  # min size of pool
     region: Optional[Region]
     scaleset_size: int  # Individual scaleset size
     spot_instances: bool = Field(default=False)
+    ephemeral_os_disks: bool = Field(default=False)
     vm_sku: str
 
-    @validator("scaleset_size", allow_reuse=True)
-    def check_scaleset_size(cls, value: int) -> int:
-        if value < 1 or value > 1000:
-            raise ValueError("invalid scaleset size")
-        return value
-
     @root_validator()
     def check_data(cls, values: Any) -> Any:
-        if (
-            "max_size" in values
-            and values.get("max_size")
-            and values.get("min_size") > values.get("max_size")
-        ):
+        if values["min_size"] <= values["max_size"]:
             raise ValueError("The pool min_size is greater than max_size")
         return values
 
-    @validator("max_size", allow_reuse=True)
-    def check_max_size(cls, value: Optional[int]) -> Optional[int]:
-        if value and value < 1:
-            raise ValueError("Autoscale sizes are not defined properly")
-        return value
-
-    @validator("min_size", allow_reuse=True)
-    def check_min_size(cls, value: int) -> int:
-        if value < 0 or value > 1000:
-            raise ValueError("Invalid pool min_size")
-        return value
-
 
 class Pool(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     name: PoolName
     pool_id: UUID = Field(default_factory=uuid4)
     os: OS
@@ -541,6 +639,7 @@ class ScalesetNodeState(BaseModel):
 
 
 class Scaleset(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     pool_name: PoolName
     scaleset_id: UUID = Field(default_factory=uuid4)
     state: ScalesetState = Field(default=ScalesetState.init)
@@ -548,26 +647,23 @@ class Scaleset(BaseModel):
     vm_sku: str
     image: str
     region: Region
-    size: int
+    size: int = Field(ge=0)
     spot_instances: bool
+    ephemeral_os_disks: bool = Field(default=False)
+    needs_config_update: bool = Field(default=False)
     error: Optional[Error]
     nodes: Optional[List[ScalesetNodeState]]
     client_id: Optional[UUID]
     client_object_id: Optional[UUID]
     tags: Dict[str, str] = Field(default_factory=lambda: {})
 
-    @validator("size", allow_reuse=True)
-    def check_size(cls, value: int) -> int:
-        if value < 0:
-            raise ValueError("Invalid scaleset size")
-        return value
-
 
 class NotificationConfig(BaseModel):
     config: NotificationTemplate
 
 
 class Repro(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     vm_id: UUID = Field(default_factory=uuid4)
     task_id: UUID
     config: ReproConfig
@@ -577,6 +673,7 @@ class Repro(BaseModel):
     error: Optional[Error]
     ip: Optional[str]
     end_time: Optional[datetime]
+    user_info: Optional[UserInfo]
 
 
 class ExitStatus(BaseModel):
@@ -662,25 +759,8 @@ class NodeEventEnvelope(BaseModel):
     event: NodeEventShim
 
 
-class StopNodeCommand(BaseModel):
-    pass
-
-
-class StopTaskNodeCommand(BaseModel):
-    task_id: UUID
-
-
-class NodeCommand(EnumModel):
-    stop: Optional[StopNodeCommand]
-    stop_task: Optional[StopTaskNodeCommand]
-
-
-class NodeCommandEnvelope(BaseModel):
-    command: NodeCommand
-    message_id: str
-
-
 class TaskEvent(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     task_id: UUID
     machine_id: UUID
     event_data: WorkerEvent
@@ -699,6 +779,7 @@ class NodeAssignment(BaseModel):
 
 
 class Task(BaseModel):
+    timestamp: Optional[datetime] = Field(alias="Timestamp")
     job_id: UUID
     task_id: UUID = Field(default_factory=uuid4)
     state: TaskState = Field(default=TaskState.init)
@@ -711,3 +792,30 @@ class Task(BaseModel):
     events: Optional[List[TaskEventSummary]]
     nodes: Optional[List[NodeAssignment]]
     user_info: Optional[UserInfo]
+
+
+class InstanceConfig(BaseModel):
+    # initial set of admins can only be set during deployment.
+    # if admins are set, only admins can update instance configs.
+    admins: Optional[List[UUID]] = None
+
+    # if set, only admins can manage pools or scalesets
+    allow_pool_management: bool = Field(default=True)
+
+    def update(self, config: "InstanceConfig") -> None:
+        for field in config.__fields__:
+            # If no admins are set, then ignore setting admins
+            if field == "admins" and self.admins is None:
+                continue
+
+            if hasattr(self, field):
+                setattr(self, field, getattr(config, field))
+
+    @validator("admins", allow_reuse=True)
+    def check_admins(cls, value: Optional[List[UUID]]) -> Optional[List[UUID]]:
+        if value is not None and len(value) == 0:
+            raise ValueError("admins must be None or contain at least one UUID")
+        return value
+
+
+_check_hotfix()
