@@ -3,11 +3,12 @@
 # Copyright (c) Microsoft Corporation.
 # Licensed under the MIT License.
 
-from typing import Dict, List, Optional
+from typing import Any, Dict, List, Optional
 from uuid import UUID
 
-from pydantic import AnyHttpUrl, BaseModel, Field, validator
+from pydantic import AnyHttpUrl, BaseModel, Field, root_validator
 
+from ._monkeypatch import _check_hotfix
 from .consts import ONE_HOUR, SEVEN_DAYS
 from .enums import (
     OS,
@@ -17,9 +18,9 @@ from .enums import (
     PoolState,
     ScalesetState,
     TaskState,
-    WebhookEventType,
 )
-from .models import AutoScaleConfig, NotificationConfig
+from .events import EventType
+from .models import AutoScaleConfig, InstanceConfig, NotificationConfig
 from .primitives import Container, PoolName, Region
 
 
@@ -57,13 +58,7 @@ class TaskSearch(BaseRequest):
 
 
 class TaskResize(TaskGet):
-    count: int
-
-    @validator("count", allow_reuse=True)
-    def check_count(cls, value: int) -> int:
-        if value <= 0:
-            raise ValueError("invalid count")
-        return value
+    count: int = Field(ge=1)
 
 
 class NodeCommandGet(BaseRequest):
@@ -107,22 +102,27 @@ class PoolStop(BaseRequest):
 
 
 class ProxyGet(BaseRequest):
-    scaleset_id: UUID
-    machine_id: UUID
-    dst_port: int
+    scaleset_id: Optional[UUID]
+    machine_id: Optional[UUID]
+    dst_port: Optional[int]
+
+    @root_validator()
+    def check_proxy_get(cls, value: Any) -> Any:
+        check_keys = ["scaleset_id", "machine_id", "dst_port"]
+        included = [x in value for x in check_keys]
+        if any(included) and not all(included):
+            raise ValueError(
+                "ProxyGet must provide all or none of the following: %s"
+                % ", ".join(check_keys)
+            )
+        return value
 
 
 class ProxyCreate(BaseRequest):
     scaleset_id: UUID
     machine_id: UUID
     dst_port: int
-    duration: int
-
-    @validator("duration", allow_reuse=True)
-    def check_duration(cls, value: int) -> int:
-        if value < ONE_HOUR or value > SEVEN_DAYS:
-            raise ValueError("invalid duration")
-        return value
+    duration: int = Field(ge=ONE_HOUR, le=SEVEN_DAYS)
 
 
 class ProxyDelete(BaseRequest):
@@ -135,7 +135,7 @@ class NodeSearch(BaseRequest):
     machine_id: Optional[UUID]
     state: Optional[List[NodeState]]
     scaleset_id: Optional[UUID]
-    pool_name: Optional[str]
+    pool_name: Optional[PoolName]
 
 
 class NodeGet(BaseRequest):
@@ -160,13 +160,7 @@ class ScalesetStop(BaseRequest):
 
 class ScalesetUpdate(BaseRequest):
     scaleset_id: UUID
-    size: Optional[int]
-
-    @validator("size", allow_reuse=True)
-    def check_optional_size(cls, value: Optional[int]) -> Optional[int]:
-        if value is not None and value < 0:
-            raise ValueError("invalid size")
-        return value
+    size: Optional[int] = Field(ge=1)
 
 
 class ScalesetCreate(BaseRequest):
@@ -174,16 +168,11 @@ class ScalesetCreate(BaseRequest):
     vm_sku: str
     image: str
     region: Optional[Region]
-    size: int
+    size: int = Field(ge=1)
     spot_instances: bool
+    ephemeral_os_disks: bool = Field(default=False)
     tags: Dict[str, str]
 
-    @validator("size", allow_reuse=True)
-    def check_size(cls, value: int) -> int:
-        if value <= 0:
-            raise ValueError("invalid size")
-        return value
-
 
 class ContainerGet(BaseRequest):
     name: Container
@@ -215,7 +204,7 @@ class CanScheduleRequest(BaseRequest):
 class WebhookCreate(BaseRequest):
     name: str
     url: AnyHttpUrl
-    event_types: List[WebhookEventType]
+    event_types: List[EventType]
     secret_token: Optional[str]
 
 
@@ -230,6 +219,18 @@ class WebhookGet(BaseModel):
 class WebhookUpdate(BaseModel):
     webhook_id: UUID
     name: Optional[str]
-    event_types: Optional[List[WebhookEventType]]
+    event_types: Optional[List[EventType]]
     url: Optional[AnyHttpUrl]
     secret_token: Optional[str]
+
+
+class NodeAddSshKey(BaseModel):
+    machine_id: UUID
+    public_key: str
+
+
+class InstanceConfigUpdate(BaseModel):
+    config: InstanceConfig
+
+
+_check_hotfix()
